      PROGRAM  twodemo
c***********************************************************************
c Copyright (C) 1993 Arve Kylling
c
c This program is free software; you can redistribute it and/or modify
c it under the terms of the GNU General Public License as published by
c the Free Software Foundation; either version 1, or (at your option)
c any later version.
c
c This program is distributed in the hope that it will be useful,
c but WITHOUT ANY WARRANTY; without even the implied warranty of
c MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the
c GNU General Public License for more details.
c
c To obtain a copy of the GNU General Public License write to the
c Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
c USA.
c***********************************************************************
c
c Run test problems for `twostr' and checks answers. The test
c problems for a solar source are taken from Toon et al., J.G.R., 
c Vol. 94, No. D13, Pages 16,287-16,301, 1989, and references
c therein. The tests for a thermal source are from KST.
c
c  KST: Kylling, A., K. Stamnes and S.-C. Tsay 1995: A reliable
c    and efficient two-stream algorithm for radiative
c    transfer; Documentation of accuracy in realistic
c    layered media, in print, Journal of Atmospheric
c    Chemistry 21, 115-150.
c
c To run the program, compile it and twostr.f, slimpak.f, and
c R1MACH.f90 or D1MACH.f90.
c
c The program produces LaTeX output that will when processed by
c LaTeX, reproduce Tables 1-5 in KST. Some editing of the output
c file before sending it to LaTeX is recommended. That is, remove
c all lines which are warnings from twostr (this has been done
c in the sample output TWODEMO.OUT, warnings should only appear 
c for Tables 4 and 5). 
c
c For documentation of variables see twostr.doc.
c
c Routines called: 
c
c    twostr: the two-stream radiative transfer program
c
c    chetwo: data block containing correct fluxes and mean
c            intensities
c   
c    ratio:  ratio of calculated to correct value with
c            underflow and overflow protection
c
c+---------------------------------------------------------------------+
c
c twostr i/o specifications **
c
      PARAMETER ( maxcly = 8, maxulv = maxcly+1, nerr=22 )
      CHARACTER  header*127
      LOGICAL  deltam, planck, prnt(2), quiet, spher, usrtau 
      INTEGER  ierror(nerr), nlyr, ntau
      REAL albedo, btemp, dtauc(maxcly), fbeam, fisot, gg(maxcly),
     $     ssalb(maxcly), radius, temis, temper(0:maxcly), ttemp, 
     $     wvnmhi, wvnmlo, umu0, utau(maxulv), zd(0:maxcly)
c
      REAL dfdt(maxulv), flup(maxulv), rfldir(maxulv),
     $     rfldn(maxulv), uavg(maxulv)
c
c Correct answers
c
      PARAMETER  ( mxprob = 9, mxcase = 27, maxtau = 7 )
c
      REAL  tstfdn( maxtau, mxcase, mxprob ),
     $      tstfup( maxtau, mxcase, mxprob ),
     $      tstdfd( maxtau, mxcase, mxprob )
      REAL  savarr(21, 11)
c
      COMMON / twoche / tstfdn, tstfup, tstdfd
c
      PARAMETER     ( mxtau = 7, mxalb=3, mxmu = 3 )
      REAL albeda(mxtau), btempa(mxtau),
     $     cmpuav( mxtau, mxtau, mxalb, mxmu ), dtauca(mxtau), 
     $     frac(mxtau), gga(mxtau), ssalba(mxtau), umu0a(mxmu)
c
      pi = 2. * ASIN( 1.0 )
      quiet = .TRUE.
      prnt(1)  = .TRUE.
      prnt(2)  = .TRUE.
c
c
c Test problem 1
c
c       References: 
c         1) KST, Table 1
c         2) Toon et al., J.G.R., Vol. 94, No. D13,
c            Pages 16,287-16,301, 1989, Table 4
c
      nlyr     = 1
      usrtau   = .TRUE.
      ntau     = 2
      utau( 1 )= 0.0
      albedo   = 0.0
      fisot    = 0.0
      spher    = .FALSE.
      deltam   = .TRUE.
      planck   = .FALSE.
c
      WRITE( header, * ) 'Test no. 1', 
     $     ':  Refs. KST, Table 1, and Toon et al., J.G.R.,',
     $     'Vol. 94, No. D13, Pages 16,287--16,301, 1989,',
     $     'Table 4.'
c
      DO 10  icas = 1, 5
         IF ( icas.EQ.1 ) THEN
            umu0       = 1.0
            utau( 2 )  = 1.0
            ssalb( 1 ) = 1.0
            gg(1)      = 0.794
            fbeam      = pi
         ELSE IF ( icas.EQ.2 ) THEN
            umu0       = 1.0
            utau( 2 )  = 1.0
            ssalb( 1 ) = 0.9
            gg(1)      = 0.794
            fbeam      = pi
         ELSE IF ( icas.EQ.3 ) THEN
            umu0       = 0.5
            utau( 2 )  = 1.0
            ssalb( 1 ) = 0.9
            gg(1)      = 0.794
            fbeam      = pi
         ELSE IF ( icas.EQ.4 ) THEN
            umu0       =  1.0
            utau( 2 )  = 64.0
            ssalb( 1 ) =  1.0
            gg(1)      =  0.848
            fbeam      = pi
         ELSE IF ( icas.EQ.5 ) THEN
            umu0       =  1.0
            utau( 2 )  = 64.0
            ssalb( 1 ) =  0.9
            gg(1)      =  0.848
            fbeam      = pi
         END IF
c
         dtauc( 1 ) = utau( 2 )
c
         CALL twostr( albedo, btemp, deltam, dtauc, fbeam, fisot,
     $        gg, header, ierror, maxcly, maxulv, nlyr, planck,
     $        ntau, prnt, quiet, radius, spher, ssalb, temis,  
     $        temper, ttemp, umu0,  usrtau, utau, wvnmlo,
     $        wvnmhi, zd, dfdt, flup, rfldir, rfldn, uavg )
c
         DO ierr = 1, nerr
            IF ( ierror(ierr) .NE. 0 ) THEN
               WRITE(*,'(/,A,I4,/)')  "TWOSTR REPORTS FATAL ERROR: ",
     $              ierr
            ENDIF
         ENDDO
c
         nprob = 1
         IF( nprob.GT.mxprob .OR. icas.GT.mxcase )  CALL
     $       ERRMSG( 'Out of bonds in exact-answer arrays', .FALSE. )
         CALL PRTFIN( albeda, albedo, cmpuav, dfdt, dtauca, flup,
     $        gg, header, icas, maxtau, maxulv, mxalb, mxmu, mxtau,
     $        ndtauc, nprob, ntau, rfldir, rfldn, savarr, ssalb,
     $        tstdfd(1, icas, nprob), tstfdn(1,1,nprob),
     $        tstfup(1, icas, nprob), uavg, umu0, umu0a, utau )
c
 10   CONTINUE
c
c
c Test problem 2
c
c       References: 
c         1) KST, Table 2
c         2) Toon et al., J.G.R., Vol. 94, No. D13,
c            Pages 16,287-16,301, 1989, Table 5
c
      nlyr      = 6
      usrtau    = .TRUE.
      ntau      = 7
      utau( 1 ) = 0.0
      albedo    = 0.0
      fisot     = 0.0
      deltam    = .TRUE.
      spher     = .FALSE.
      planck    = .FALSE.
      frac(2)   = 20.
      frac(3)   = 10.
      frac(4)   =  5.
      frac(5)   =  2.
      frac(6)   =  4./3.
      frac(7)   =  1.
c
      DO 20  icas = 1, 5
      WRITE( header, * ) 'Test no. 2', 
     $     ':  Refs. KST, Table 2, and Toon et al., J.G.R.,',
     $     'Vol. 94, No. D13, Pages 16,287--16,301, 1989,',
     $     'Table 5.'
         IF ( icas.EQ.1 ) THEN
            umu0       = 1.0
            tautot     = 1.0
            DO 21 lc = 1, ntau
               IF (lc.GE.2) utau(lc)  = tautot/frac(lc)
               ssalb(lc) = 1.0
               gg(lc)    = 0.794
 21         CONTINUE
            fbeam      = pi
         ELSE IF ( icas.EQ.2 ) THEN
            umu0       = 1.0
            tautot     = 1.0
            DO 22 lc = 1, ntau
               IF (lc.GE.2) utau(lc)  = tautot/frac(lc)
               ssalb(lc) = 0.9
               gg(lc)    = 0.794
 22         CONTINUE
            fbeam      = pi
         ELSE IF ( icas.EQ.3 ) THEN
            umu0       = 0.5
            tautot     = 1.0
            DO 23 lc = 1, ntau
               IF (lc .GE. 2) utau(lc)  = tautot/frac(lc)
               ssalb(lc) = 0.9
               gg(lc)    = 0.794
 23         CONTINUE
            fbeam      = pi
         ELSE IF ( icas.EQ.4 ) THEN
            umu0       =  1.0
            tautot     = 64.0
            DO 24 lc = 1, ntau
               IF (lc.GE.2) utau(lc)  = tautot/frac(lc)
               ssalb(lc) = 1.0
               gg(lc)     = 0.848
 24         CONTINUE
            fbeam      = pi
         ELSE IF ( icas.EQ.5 ) THEN
            umu0       =  1.0
            tautot     = 64.0
            DO 25 lc = 1, ntau
               IF (lc.GE.2) utau(lc)  = tautot/frac(lc)
               ssalb(lc) = 0.9
               gg(lc)    = 0.848
 25         CONTINUE
            fbeam      = pi
         END IF
c
         taut     = 0.0
         lc       = 1
         dtauc(1) = utau(2) - utau(1)
         taut     = taut + dtauc(1)
         DO 26 lc = 2, nlyr
            lu        = lc 
            dtauc(lc) = utau(lu+1) -utau(lu)
            taut      = taut + dtauc(lc)
 26      CONTINUE
c
         CALL twostr( albedo, btemp, deltam, dtauc, fbeam, fisot,  
     $        gg, header, ierror, maxcly, maxulv, nlyr, planck,
     $        ntau, prnt, quiet, radius, spher, ssalb, temis,  
     $        temper, ttemp, umu0,  usrtau, utau, wvnmlo,
     $        wvnmhi, zd, dfdt, flup, rfldir, rfldn, uavg )
c
         DO ierr = 1, nerr
            IF ( ierror(ierr) .NE. 0 ) THEN
               WRITE(*,'(/,A,I4,/)')  "TWOSTR REPORTS FATAL ERROR: ",
     $              ierr
            ENDIF
         ENDDO
c
         nprob = 2
         ls = 1
         DO 200 lu = 1, ntau
            fdntot = rfldn( lu ) + rfldir( lu )
            fnet   = fdntot - flup( lu )
            fdiv   = 0.0
            IF ( lu .LT. ntau ) THEN
               fdntot1 = rfldn( lu+1 ) + rfldir( lu+1 )
               fnet1   = fdntot1 - flup( lu+1 )
               fdiv = fnet - fnet1
            ENDIF
            ic = (icas - 1 )*2 + 2
            savarr(ls, 1 ) = utau(lu)
            savarr(ls, ic ) = tstfdn(lu,icas,nprob)
            savarr(ls, ic +1) = tstfup(lu,icas,nprob)
            ls = ls + 1
            savarr(ls, 1 ) = utau(lu)
            savarr(ls, ic ) = fnet
            savarr(ls, ic + 1 ) = fdiv
            ls = ls + 1
            savarr(ls, 1 ) = utau(lu)
            savarr(ls, ic ) =
     $           -(ratio(fnet,tstfdn(lu,icas,nprob))-1.)*100.0
            savarr(ls, ic+1) =
     $           -(ratio(fdiv,tstfup(lu,icas,nprob))-1.)*100.0
            ls = ls + 1
 200     CONTINUE
 20   CONTINUE
      icas = 5
      IF( nprob.GT.mxprob .OR. icas.GT.mxcase )  CALL
     $     ERRMSG( 'Out of bonds in exact-answer arrays', .FALSE. )
      CALL PRTFIN( albeda, albedo, cmpuav, dfdt, dtauca, flup,
     $     gg, header, icas, maxtau, maxulv, mxalb, mxmu, mxtau,
     $     ndtauc, nprob, ntau, rfldir, rfldn, savarr, ssalb,
     $     tstdfd(1, icas, nprob), tstfdn(1,1,nprob),
     $     tstfup(1, icas, nprob), uavg, umu0, umu0a, utau )
c
c     
c Test problem 3
c
c       References: 
c         1) KST, Table 3
c         2) Toon et al., J.G.R., Vol. 94, No. D13,
c            Pages 16,287-16,301, 1989, Table 6
c
      nlyr      = 1
      usrtau    = .TRUE.
      ntau      = 2
      utau( 1 ) = 0.0
      albedo    = 0.0
      fisot     = 0.0
      deltam    = .TRUE.
      planck    = .FALSE.
      umu0a(1)  = 0.1
      umu0a(2)  = 0.4
      umu0a(3)  = 0.92
      dtauca(1) = 0.02
      dtauca(2) = 0.25
      dtauca(3) = 1.0
      albeda(1) = 0.0
      albeda(2) = 0.25
      albeda(3) = 0.8
c
      numu0     = 3
      ndtauc    = 3
      nalb      = 3
      icas      = 0
c
      DO iumu0 = 1, numu0
         DO idtauc = 1, ndtauc
            DO ialb = 1, nalb
               icas = icas + 1
c
      WRITE( header, * ) 'Test no. 3', 
     $     ':  Refs. KST, Table 3, and Toon et al., J.G.R.,',
     $     'Vol. 94, No. D13, Pages 16,287--16,301, 1989,',
     $     'Table 6.'
               umu0       = umu0a(iumu0)
               utau( 2 )  = dtauca(idtauc)
               dtauc( 1 ) = dtauca(idtauc)
               albedo     = albeda(ialb)
               ssalb( 1 ) = 1.0
               gg(1)      = 0.0
               fbeam      = 1.0
c
               CALL twostr( albedo, btemp, deltam, dtauc, fbeam, fisot,
     $              gg, header, ierror, maxcly, maxulv, nlyr, planck,
     $              ntau, prnt, quiet, radius, spher, ssalb, temis,  
     $              temper, ttemp, umu0,  usrtau, utau, wvnmlo,
     $              wvnmhi, zd, dfdt, flup, rfldir, rfldn, uavg )
c
         DO ierr = 1, nerr
            IF ( ierror(ierr) .NE. 0 ) THEN
               WRITE(*,'(/,A,I4,/)')  "TWOSTR REPORTS FATAL ERROR: ",
     $              ierr
            ENDIF
         ENDDO
c
c
               cmpuav(1, idtauc, ialb, iumu0) = uavg(1)
               cmpuav(2, idtauc, ialb, iumu0) = uavg(2)
c
               nprob = 3
            ENDDO
         ENDDO
      ENDDO
      IF( nprob.GT.mxprob .OR. icas.GT.mxcase )  CALL
     $     ERRMSG( 'Out of bonds in exact-answer arrays', .FALSE. )
      CALL PRTFIN( albeda, albedo, cmpuav, dfdt, dtauca, flup,
     $     gg, header, icas, maxtau, maxulv, mxalb, mxmu, mxtau,
     $     ndtauc, nprob, ntau, rfldir, rfldn, savarr, ssalb,
     $     tstdfd(1, icas, nprob), tstfdn(1,1,nprob),
     $     tstfup(1, icas, nprob), uavg, umu0, umu0a, utau )
c
c
c Test problem 4
c
c       References: 
c         1) KST, Table 4
c
      nlyr = 1
      usrtau = .TRUE.
      ntau = 2
      utau( 1 ) = 0.0
      albedo = 0.0
      fisot = 0.0
      deltam = .TRUE.
      planck = .TRUE.
      dtauca(1) = 1.0
      dtauca(2) = 1000.0
      dtauca(3) = 10.
      fbeam     = 0.0
      gg(1)      = 0.0
      ssalba(1) = 0.0
      ssalba(2) = 0.0
      ssalba(3) = 1.0
      temper(0) = 270.0
      temper(1) = 280.0
      btempa(1)  = 0.0
      btempa(2)  = 0.0
      btempa(3)  = 300.0
      temis     = 0.0
      wvnmlo    = 0.0
      wvnmhi    = 10000.0
c
      ndtauc = 3
      icas   = 0
      DO idtauc = 1, ndtauc
         icas = icas + 1
         WRITE( header, * ) 'Test no. 4:  Refs. KST, Table 4'
         utau( 2 )  = dtauca(idtauc)
         dtauc( 1 ) = dtauca(idtauc)
         ssalb( 1 ) = ssalba(icas)
         btemp      = btempa(icas)
c     
         if ( idtauc .eq. ndtauc ) temper(0) = temper(1)
         CALL twostr( albedo, btemp, deltam, dtauc, fbeam, fisot,
     $        gg, header, ierror, maxcly, maxulv, nlyr, planck,
     $        ntau, prnt, quiet, radius, spher, ssalb, temis,  
     $        temper, ttemp, umu0,  usrtau, utau, wvnmlo,
     $        wvnmhi, zd, dfdt, flup, rfldir, rfldn, uavg )
c     
c
         DO ierr = 1, nerr
            IF ( ierror(ierr) .NE. 0 ) THEN
               WRITE(*,'(/,A,I4,/)')  "TWOSTR REPORTS FATAL ERROR: ",
     $              ierr
            ENDIF
         ENDDO
c
         nprob = 4
c     
c     Save output and Exact answers in sav(e)arr(ay) to be handed
c     to prtfin. (This is a little akward, but so is Table 4, and
c     this works!!)
c
	 savarr(icas,1)   = dtauca(idtauc)
	 savarr(icas+3,1) = dtauca(idtauc)
	 savarr(icas,2)   = ssalba(idtauc)
	 savarr(icas+3,2) = ssalba(idtauc)
	 savarr(icas,3)   = tstfup(1, icas, nprob)
	 savarr(icas+3,3) = tstdfd(1, icas, nprob)
	 savarr(icas,4)   = flup(1)
	 savarr(icas+3,4) = dfdt(1)
	 savarr(icas,5)   = 
     $         (ratio(flup(1),tstfup(1,icas,nprob))-1.0)*100.0
	 savarr(icas+3,5) = 
     $         (ratio(dfdt(1),tstdfd(1,icas,nprob))-1.0)*100.0
	 savarr(icas,6)   = tstfdn(1, icas, nprob)
	 savarr(icas+3,6) = tstdfd(ntau, icas, nprob)
	 savarr(icas,7)   = rfldn(ntau)
	 savarr(icas+3,7) = dfdt(ntau)
	 savarr(icas,8)   = 
     $         (ratio(rfldn(ntau),tstfdn(1,icas,nprob))-1.0)*100.0
	 savarr(icas+3,8) = 
     $         (ratio(dfdt(ntau),tstdfd(ntau,icas,nprob))-1.0)*100.0
         IF( nprob.GT.mxprob .OR. icas.GT.mxcase )  CALL
     $        ERRMSG( 'Out of bonds in exact-answer arrays',
     $        .FALSE. )
      ENDDO
      CALL PRTFIN( albeda, albedo, cmpuav, dfdt, dtauca, flup,
     $     gg, header, icas, maxtau, maxulv, mxalb, mxmu, mxtau,
     $     ndtauc, nprob, ntau, rfldir, rfldn, savarr, ssalb,
     $     tstdfd(1, icas, nprob), tstfdn(1,1,nprob),
     $     tstfup(1, icas, nprob), uavg, umu0, umu0a, utau )
c
c
c Test problem 5
c
c       References: 
c         1) KST, Table 5
c
      nlyr = 1
      usrtau = .TRUE.
      ntau = 2
      utau( 1 ) = 0.0
      albedo = 0.0
      fisot = 0.0
      deltam = .TRUE.
      planck = .TRUE.
      dtauca(1) = 0.1
      dtauca(2) = 1.0
      dtauca(3) = 10.0
      dtauca(4) = 100.0
      fbeam     = 0.0
      ssalba(1) = 0.1
      ssalba(2) = 0.95
      temper(0) = 200.0
      temper(1) = 300.0
      btemp     = 0.0
      wvnmlo    = 300.0
      wvnmhi    = 800.0
      gga(1)    = 0.05
      gga(2)    = 0.75
c
      ndtauc = 4
      ngg       = 2
      icas = 0
      DO idtauc = 1, ndtauc
         DO igg = 1, ngg
            icas = icas + 1
            WRITE( header, * ) 'Test no. 5:  Refs. KST, Table 1'
            utau( 2 )  = dtauca(idtauc)
            dtauc( 1 ) = dtauca(idtauc)
            ssalb( 1 ) = ssalba(igg)
            gg(1)      = gga(igg)
c     
            CALL twostr( albedo, btemp, deltam, dtauc, fbeam, fisot,
     $           gg, header, ierror, maxcly, maxulv, nlyr, planck,
     $           ntau, prnt, quiet, radius, spher, ssalb, temis,  
     $           temper, ttemp, umu0,  usrtau, utau, wvnmlo,
     $           wvnmhi, zd, dfdt, flup, rfldir, rfldn, uavg )
c
            DO ierr = 1, nerr
               IF ( ierror(ierr) .NE. 0 ) THEN
                  WRITE(*,'(/,A,I4,/)')  "TWOSTR REPORTS FATAL ERROR: ",
        $              ierr
               ENDIF
            ENDDO
c
c
            nprob = 5
            IF( nprob.GT.mxprob .OR. icas.GT.mxcase )  CALL
     $           ERRMSG( 'Out of bonds in exact-answer arrays',
     $           .FALSE. )
            CALL PRTFIN( albeda, albedo, cmpuav, dfdt, dtauca, flup,
     $           gg, header, icas, maxtau, maxulv, mxalb,mxmu,mxtau,
     $           ndtauc, nprob, ntau, rfldir, rfldn, savarr, ssalb,
     $           tstdfd(1, icas, nprob), tstfdn(1,1,nprob),
     $           tstfup(1, icas, nprob), uavg, umu0, umu0a, utau )
         ENDDO
      ENDDO
      
      STOP
      END

      SUBROUTINE PRTFIN( albeda, albedo, cmpuav, dfdt, dtauca, flup,
     $     gg, header, icas, maxtau, maxulv, mxalb, mxmu, mxtau, 
     $     ndtauc, nprob, ntau, rfldir, rfldn, savarr, ssalb, tstdfd, 
     $     tstfdn, tstfup, uavg, umu0, umu0a, utau )
c
      CHARACTER header*127
      REAL albeda(*), cmpuav(mxtau, mxtau, mxalb, mxmu ), dfdt(*),
     $     dtauca(*), flup(*), gg(*), rfldir(*), rfldn(*),
     $     savarr(21,*),ssalb(*), tstdfd(*), tstfdn(maxtau, *),
     $     tstfup(*), uavg(*),umu0a(*), utau(*)
c
      pi = 2. * ASIN( 1.0 )
      IF ( ntau.GT.maxtau ) CALL
     $     ERRMSG('prtfin--out of bounds in comparator arrays',.TRUE.)
      WRITE(*,1010)
c
      IF ( nprob .EQ. 1 ) THEN

         WRITE(*,110) icas, umu0, utau(2), ssalb(1), gg(1),
     $        tstfup(1), flup(1),  (ratio(flup(1),tstfup(1))-1.)*100.,
     $        tstfdn(2,icas), rfldn(2),
     $        (ratio(rfldn(2),tstfdn(2,icas))-1.)*100. 
         IF ( icas .EQ. 5 )  WRITE(*,199) 
c
      ELSE IF ( nprob .EQ. 2 ) THEN
      
         WRITE(*,203)(ic,ic = 1,5)
         DO ls = 1, 2
            WRITE(*,211) (savarr(ls, ic), ic = 2,11 )
         ENDDO
         ls = 3
         WRITE(*,212) (savarr(ls, ic), ic = 2,11 )
         DO ls = 4, 5
            WRITE(*,213) (savarr(ls, ic), ic = 2,11 )
         ENDDO
         ls = 6
         WRITE(*,214) (savarr(ls, ic), ic = 2,11 )
         DO ls = 7, 8
            WRITE(*,215) (savarr(ls, ic), ic = 2,11 )
         ENDDO
         ls = 9
         WRITE(*,216) (savarr(ls, ic), ic = 2,11 )
         DO ls = 10, 11
            WRITE(*,217) (savarr(ls, ic), ic = 2,11 )
         ENDDO
         ls = 12
         WRITE(*,218) (savarr(ls, ic), ic = 2,11 )
         DO ls = 13, 14
            WRITE(*,219) (savarr(ls, ic), ic = 2,11 )
         ENDDO
         ls = 15
         WRITE(*,220) (savarr(ls, ic), ic = 2,11 )
         DO ls = 16, 17
            WRITE(*,221) (savarr(ls, ic), ic = 2,11 )
         ENDDO
         ls = 18
         WRITE(*,222) (savarr(ls, ic), ic = 2,11 )
         DO ls = 19, 20
            WRITE(*,223) (savarr(ls, ic), ic = 2,11 )
         ENDDO
         ls = 21
         WRITE(*,224) (savarr(ls, ic), ic = 2,11 )
         WRITE(*,299) 
c
      ELSE IF ( nprob .EQ. 3 ) THEN
      
         WRITE(*,310)(albeda(ialb),ialb=1,3)
         DO 302 lu = 1, ntau
            icase = 0
            IF ( lu .EQ. 2) WRITE(*,312)
            DO 301 iumu0 = 1, 3
               DO 300 lev = 1, ndtauc
                  icase = icase + 1
                  uavgex0 = tstfdn(lu, icase)
                  icase = icase + 1
                  uavgex1 = tstfdn(lu, icase)
                  icase = icase + 1
                  uavgex2 = tstfdn(lu, icase)
                  DO ialb = 1, 3
                     cmpuav(lu, lev, ialb, iumu0) =
     $                    4.0*pi*cmpuav(lu, lev, ialb, iumu0)
                  ENDDO
                  WRITE(*,311) umu0a(iumu0), dtauca(lev), 
     $                 uavgex0, cmpuav(lu, lev, 1, iumu0),
     $                 (ratio(cmpuav(lu, lev, 1, iumu0),
     $                 uavgex0)-1.)*100., 
     $                 uavgex1, cmpuav(lu, lev, 2, iumu0),
     $                 (ratio(cmpuav(lu, lev, 2, iumu0),
     $                 uavgex1)-1.)*100.,
     $                 uavgex2, cmpuav(lu, lev, 3, iumu0),
     $                 (ratio(cmpuav(lu, lev, 3, iumu0),
     $                 uavgex2)-1.)*100. 
 300           CONTINUE
 301        CONTINUE
 302     CONTINUE
         WRITE(*,399)
c
      ELSE IF ( nprob .EQ. 4 ) THEN
      
         WRITE(*,412) 
     $            (savarr(1,ic), ic = 1,8),
     $            (savarr(2,ic), ic = 1,8)
         WRITE(*,413)  (savarr(3,ic), ic = 1,8)
         WRITE(*,412) 
     $            (savarr(4,ic), ic = 1,8),
     $            (savarr(5,ic), ic = 1,8)
         WRITE(*,413)  (savarr(6,ic), ic = 1,8)
         WRITE(*,499) 
c
      ELSE IF ( nprob .EQ. 5 ) THEN

         topnet = ( rfldn(1) + rfldir(1) ) - flup(1) 
         botnet = ( rfldn(ntau) + rfldir(ntau) ) - flup(ntau)
         botdwn = rfldn(ntau) + rfldir(ntau)
         WRITE(*,511) utau(2), ssalb(1), gg(1), 
     $        tstfup(1),flup(1),  (ratio(flup(1),tstfup(1))-1.0)*100.0,
     $        tstdfd(1), botdwn,  (ratio(botdwn,tstdfd(1))-1.0)*100.0,
     $        tstfdn(1,icas), topnet - botnet, 
     $        (ratio(topnet-botnet,tstfdn(1,icas))-1.0)*100.0
         IF ( icas .EQ. 8 )  WRITE(*,599) 
         
      ENDIF
c
 1010 FORMAT ( ////,  Two stream method radiative ',
     $  'transfer test program output',/,
     $  'For more information, see Kylling, Stamnes and Tsay, ',/,
     $  '1995, J. Atmospheric Chemistry.')

 199  FORMAT('Single layer results for the diffuse upward ',/,
     $     'and downward fluxes at the top and bottom of the layer.',/,
     $     'The surface albedo A_g=0.0 for all five cases. The ',/,
     $     'exact values are from Wiscombe (1977a) and Lenoble',/,
     $     '(1985), while Twostr refers to the present method.', )

 299  FORMAT('Six layers results for the net flux, ',/,
     $     'F_net=(F+)-(F-), and the flux divergence, ',/,
     $     'F_div=F_net(layer i+1) - F_net(layer i)',/,
     $     'The first value for',/,
     $     'each tau--level is exact, the second the Twostr',/, 
     $     'results, and in parenthesis is the percent difference.',/,
     $     ' Cases and tau are defined in Table ',
     $     'ref{tab1}. The surface albedo ',/,
     $     'A_g=0.0 for all five cases. The exact values are ',/,
     $     'from Wiscombe (1977a) and Lenoble (1985), while Twostr',/,
     $     'refers to the present method.}')

 399  FORMAT('Single layer results for the mean ',/,
     $     'intensity for conservative Rayleigh scattering ',/,
     $     '(a=1, g=0). The exact values are from ',/,
     $     'Wiscombe (1977a) and Lenoble (1985), while Twostr ',/,
     $     'refers to the present method.')

 499  FORMAT('Upward and downward fluxes and the flux ',/,
     $     'divergence for a single layer in limit the a=0 ',/,
     $     'and a=1. The temperature at the top of the ',/,
     $     'layer is 270.0K and 280.0K at the bottom. ',/,
     $     'It is assumed to vary linearly across the layer. ',
     $     'The surface temperature is 0.0K  for the a=0 '/, 
     $     'cases and 300.0K for the a=1 case. ',/,
     $     'The Planck function is integrated over the interval ',/,
     $     '0.0--10,000.0 cm^{-1}. Exact results ',/,
     $     'are from 16--stream calculations by the DISORT ',/,
     $     'algorithm. The asymmetry factor g=0.0}')

 599  FORMAT('Single layer results for thermal radiation ',/, 
     $     'for different optical depths, single scattering ',/,
     $     'albedos and asymmetry factors. The temperature varies ',/,
     $     'linearly between 200.0K at the top  and ',/,
     $     '300.0K at the bottom of the layer. ',/,
     $     'The Planck function is integrated over the interval ',/,
     $     '300.0--800.0 cm^{-1}. ',/,
     $     'The surface temperature is 0.0K. The exact results ',/,
     $     'are from Kylling and Stamnes (1992) and from ',/,
     $     '16--stream calculations by the DISORT algorithm.')
     
      RETURN
      END
c
      REAL FUNCTION  RATIO( A, B )
c
c        CALCULATE RATIO  A/B  WITH OVER- AND UNDER-FLOW PROTECTION
c
         IF ( ABS(A).LT.1.0E-4 .AND. ABS(B).LT.1.0E-4 )  THEN
            RATIO = 1.0
         ELSE IF ( B.EQ.0.0 )  THEN
            RATIO = 1.E+20
         ELSE
            RATIO = A / B
         END IF
c
      RETURN
      END
c
      BLOCK DATA chetwo
c
c Correct results from Toon et al, J.G.R., Vol. 94, No. D13,
c Pages 16,287--16,301, 1989, and KST
c
      PARAMETER  ( mxprob = 9, mxcase = 27, maxtau = 7 )
c
      REAL  tstfdn( maxtau, mxcase, mxprob ),
     $      tstfup( maxtau, mxcase, mxprob ),
     $      tstdfd( maxtau, mxcase, mxprob )
c
      COMMON / twoche / tstfdn, tstfup, tstdfd
c
c Test 1, case 1
c
      DATA ( tstfdn( I,1,1 ), I = 1, 2 ) / 0.0        , 1.813 /
      DATA ( tstfup( I,1,1 ), I = 1, 2 ) / 0.173      , 0.0   /
c
c Test 1, case 2
c
      DATA ( tstfdn( I,2,1 ), I = 1, 2 ) / 0.0        , 1.516 /
      DATA ( tstfup( I,2,1 ), I = 1, 2 ) / 0.124      , 0.0 /
c
c Test 1, case 3
c
      DATA ( tstfdn( I,3,1 ), I = 1, 2 ) / 0.0        , 0.803 /
      DATA ( tstfup( I,3,1 ), I = 1, 2 ) / 0.226      , 0.0 /
c
c Test 1, case 4
c
      DATA ( tstfdn( I,4,1 ), I = 1, 2 ) / 0.0        , 0.480 /
      DATA ( tstfup( I,4,1 ), I = 1, 2 ) / 2.662      , 0.0 /
c
c Test 1, case 5
c
      DATA ( tstfdn( I,5,1 ), I = 1, 2 ) / 0.0        , 0.0 /
      DATA ( tstfup( I,5,1 ), I = 1, 2 ) / 0.376      , 0.0 /
c
c For all cases in test 2 tstfdn is the net flux and tstfup the flux
c divergence
c
c Test 2, case 1
c
      DATA ( tstfdn(I,1,2), I=1,7) 
     $     /2.968, 2.968, 2.968, 2.968, 2.968, 2.968, 2.968/  
      DATA ( tstfup(I,1,2), I=1,7) 
     $     /0.0  , 0.0  , 0.0  , 0.0  , 0.0  , 0.0  , 0.0  /
c
c Test 2, case 2
c
      DATA ( tstfdn(I,2,2), I=1,7) 
     $     /3.018, 3.001, 2.983, 2.948, 2.842, 2.755, 2.671/  
      DATA ( tstfup(I,2,2), I=1,7)
     $     /0.017 ,0.018 ,0.035, 0.1060,0.0870,0.0840,0.0  /
c
c Test 2, case 3
c
      DATA ( tstfdn(I,3,2), I=1,7) 
     $     /1.345, 1.326, 1.306, 1.266, 1.157, 1.079, 1.016/  
      DATA ( tstfup(I,3,2), I=1,7) 
     $     /0.019, 0.020, 0.040, 0.109, 0.078, 0.063, 0.0  /
c
c Test 2, case 4
c
      DATA ( tstfdn(I,4,2), I=1,7) 
     $     /0.480, 0.480, 0.480, 0.480, 0.480, 0.480, 0.480/  
      DATA ( tstfup(I,4,2), I=1,7)
     $     /0.0  , 0.0  , 0.0  , 0.0  , 0.0  , 0.0  , 0.0  /
c
c Test 2, case 5
c
      DATA ( tstfdn(I,5,2), I=1,7) 
     $     /2.776, 1.605, 0.830, 0.196, 0.0022,0.0001,0.0  /  
      DATA ( tstfup(I,5,2), I=1,7)
     $     /1.161, 0.775, 0.634, 0.194,0.0021, 0.0001,0.0  /
c
c For all cases in test 3 tstfdn is the mean intensity (uavg)
c
c Test 3, case 1
c
      DATA ( tstfdn(I,1,3), I=1,2)  /1.045, 0.864 /
c
c Test 3, case 2
c
      DATA ( tstfdn(I,2,3), I=1,2)  /1.089, 0.912 /
c
c Test 3, case 3
c
      DATA ( tstfdn(I,3,3), I=1,2)  /1.187, 1.018 /
c
c Test 3, case 4
c
      DATA ( tstfdn(I,4,3), I=1,2)  /1.170, 0.192 /
c
c Test 3, case 5
c
      DATA ( tstfdn(I,5,3), I=1,2)  /1.189, 0.224 /
c
c Test 3, case 6
c
      DATA ( tstfdn(I,6,3), I=1,2)  /1.239, 0.307 /
c
c Test 3, case 7
c
      DATA ( tstfdn(I,7,3), I=1,2)  /1.212, 0.057 /
c
c Test 3, case 8
c
      DATA ( tstfdn(I,8,3), I=1,2)  /1.220, 0.082 /
c
c Test 3, case 9
c
      DATA ( tstfdn(I,9,3), I=1,2)  /1.247, 0.168 /
c
c Test 3, case 10
c
      DATA ( tstfdn(I,10,3), I=1,2)  /1.047, 0.998 /
c
c Test 3, case 11
c
      DATA ( tstfdn(I,11,3), I=1,2)  /1.235, 1.203 /
c
c Test 3, case 12
c
      DATA ( tstfdn(I,12,3), I=1,2)  /1.653, 1.661 /
c
c Test 3, case 13
c
      DATA ( tstfdn(I,13,3), I=1,2)  /1.284, 0.787 /
c
c Test 3, case 14
c
      DATA ( tstfdn(I,14,3), I=1,2)  /1.402, 0.988 /
c
c Test 3, case 15
c
      DATA ( tstfdn(I,15,3), I=1,2)  /1.707, 1.502 /
c
c Test 3, case 16
c
      DATA ( tstfdn(I,16,3), I=1,2)  /1.534, 0.385 /
c
c Test 3, case 17
c
      DATA ( tstfdn(I,17,3), I=1,2)  /1.584, 0.540 /
c
c Test 3, case 18
c
      DATA ( tstfdn(I,18,3), I=1,2)  /1.778, 1.071 /
c
c Test 3, case 19
c
      DATA ( tstfdn(I,19,3), I=1,2)  /1.040, 1.018 /
c
c Test 3, case 20
c
      DATA ( tstfdn(I,20,3), I=1,2)  /1.477, 1.495 /
c
c Test 3, case 21
c
      DATA ( tstfdn(I,21,3), I=1,2)  /2.453, 2.561 /
c
c Test 3, case 22
c
      DATA ( tstfdn(I,22,3), I=1,2)  /1.279, 1.028 /
c
c Test 3, case 23
c
      DATA ( tstfdn(I,23,3), I=1,2)  /1.597, 1.560 /
c
c Test 3, case 24
c
      DATA ( tstfdn(I,24,3), I=1,2)  /2.404, 2.928 /
c
c Test 3, case 25
c
      DATA ( tstfdn(I,25,3), I=1,2)  /1.691, 0.881 /
c
c Test 3, case 26
c
      DATA ( tstfdn(I,26,3), I=1,2)  /1.851, 1.384 /
c
c Test 3, case 27
c
      DATA ( tstfdn(I,27,3), I=1,2)  /2.398, 3.109 /
c
c Test 4, case 1
c
      DATA  tstfup( 1,1,4 ) /  248.21828 /
      DATA  tstfdn( 1,1,4 ) /  259.08502 /
      DATA ( tstdfd(I,1,4), I=1,2) /-669.36536, -823.39050 /
c
c Test 4, case 2
c
      DATA  tstfup( 1,2,4 ) /  301.37552 /
      DATA  tstfdn( 1,2,4 ) /  348.49817 /
      DATA ( tstdfd(I,2,4), I=1,2) /-602.64093, -697.10645 /
c
c Test 4, case 3
c
      DATA  tstfup( 1,3,4 ) /  53.62256 /
      DATA  tstfdn( 1,3,4 ) / 405.67416 /
      DATA ( tstdfd(I,3,4), I=1,2) /-0.00, -0.00 /
c
c For all cases in test 5 tstfdn is the net flux top - net flux bottom
c
c Test 5, case 1
c
      DATA  tstfdn( 1,1,5 ) / -39.839 /
      DATA  tstfup( 1,1,5 ) /  19.271 /
      DATA  tstdfd( 1,1,5 ) /  20.570 /
c
c Test 5, case 2
c
      DATA  tstfdn( 1,2,5 ) /  -2.575 /
      DATA  tstfup( 1,2,5 ) /   1.264 /
      DATA  tstdfd( 1,2,5 ) /   1.282 /
c
c Test 5, case 3
c
      DATA  tstfdn( 1,3,5 ) /-195.348 /
      DATA  tstfup( 1,3,5 ) /  80.164 /
      DATA  tstdfd( 1,3,5 ) / 115.189 /
c
c Test 5, case 4
c
      DATA  tstfdn( 1,4,5 ) / -24.135 /
      DATA  tstfup( 1,4,5 ) /  11.317 /
      DATA  tstdfd( 1,4,5 ) /  12.816 /
c
c Test 5, case 5
c
      DATA  tstfdn( 1,5,5 ) /-265.788 /
      DATA  tstfup( 1,5,5 ) /  63.725 /
      DATA  tstdfd( 1,5,5 ) / 202.096 /
c
c Test 5, case 6
c
      DATA  tstfdn( 1,6,5 ) /-136.632 /
      DATA  tstfup( 1,6,5 ) /  53.001 /
      DATA  tstdfd( 1,6,5 ) /  83.633 /
c
c Test 5, case 7
c
      DATA  tstfdn( 1,7,5 ) /-270.983 /
      DATA  tstfup( 1,7,5 ) /  56.541 /
      DATA  tstdfd( 1,7,5 ) / 214.720 /
c
c Test 5, case 8
c
      DATA  tstfdn( 1,8,5 ) /-172.113 /
      DATA  tstfup( 1,8,5 ) /  39.423 /
      DATA  tstdfd( 1,8,5 ) / 132.715 /
      
      END

